#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <sstream>
#include <cerrno>

#include <getopt.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <conductor.hh>

#include <command_io.hh>
#include <bus_io.hh>
#include <utils/line_iterator.hh>
#include <utils/str.hh>
#include <utils/dirstream.hh>

std::string plugin_dir="samples/";

struct append_options
{
    append_options(command& cmd): cmd_(cmd) {}
    void operator()(const std::string& line)
    {
	std::string::size_type pos = line.find(":");
	cmd_(line.substr(0,pos),line.substr(pos+1,std::string::npos));
    }
    command &cmd_;
};

/**
 * @todo it's possible we need some encoding
 */
struct dump_options
{
    dump_options(std::ostream& os): os_(os) {}
    void operator()(const std::string& key,const std::string& value)
    {
	if ("action" != key) os_<<key<<":"<<value<<std::endl;
    }
    std::ostream &os_;
};

/** print program usage information */
void usage(int retcode)
{
	extern const char *__progname;
	std::cout << "Usage: " << __progname << " [-hvra] [--help] [--version] [--admfs-path <path>]" << std::endl;
	std::cout << "Alterator module between brook and amdfs" << std::endl;
	std::cout << "Convert HOO commands to operations on admfs filesystems and inversely" << std::endl;
	std::cout << "\t-h, --help             display help screen" << std::endl;
	std::cout << "\t-v, --version          display version information" << std::endl;
	std::cout << "\t-p, --plugin-path       path to directory with admfs mounted in it" << std::endl;
	std::cout << std::endl;
	std::cout << "Report bugs to <inger@altlinux.org>" << std::endl;
	exit(retcode);
}

/* write options to the file */
bool wr(const std::string& path,command& cmd)
{
	conductor_impl& conduct = conductor::instance();
	if (!conduct.open(path.c_str(),O_WRONLY)) return false;

	std::ostringstream os(path.c_str());
	cmd.for_each(dump_options(os));

	conduct.write(path,os.str().c_str(),os.str().length(),0);
	conduct.close(path);

	return true;
}


/* read from the file and convert to options */
bool rd(const std::string&path,command& cmd)
{
	conductor_impl& conduct = conductor::instance();
	if (!conduct.open(path.c_str(),O_RDONLY)) return false;

	std::vector<char> buffer(BUFSIZ);
	std::string result;
	off_t offset = 0;
	size_t len = 0;
	while ((len=conduct.read(path,&buffer[0],buffer.size(),offset))!=0)
	{
		result+=&buffer[0];
		offset += len;
	}
	std::istringstream is(result);
	std::for_each(alt::istream_line_iterator(is),
		      alt::istream_line_iterator(),
		      append_options(cmd));

	conduct.close(path);
	return true;
}

void sure_plugin(const std::string& path)
{
	if (conductor::instance().type(path).empty())
		conductor::instance().add_plugin(path,plugin_dir);
}

struct out_cmd
{
	out_cmd(std::ostream& os,const std::string& prefix):
		os_(os), prefix_(prefix)
	{}
	void operator()(const std::string& object)
	{
		os_<<command(prefix_+"/"+object);
	}
	std::ostream&		os_;
	const std::string&	prefix_;
};


void forward(std::istream& in,std::ostream& out)
{

	conductor_impl& conduct = conductor::instance();
	
	command cmd;
	while(in>>cmd)
	{
		//ignore control commands
		if (cmd.name().empty() || alt::starts_with(cmd.name(),"/ctrl/"))
		{
			out<<command("");
			continue;
		}

		command error("/error"+cmd.name());

		const std::string path=cmd.name();
		const std::string action = cmd("action");

		error("action",action);

		//don't allow relative directories in result path
		if (path[0] != '/' || path.find("/../") != std::string::npos)
		{
			out<<error("reason","relative path detected");
			continue;
		}
	
		const std::string directory=
			path.substr(0,path.find("/",
						(path[0] == '/')?path.find_first_not_of("/"):0));

		sure_plugin(directory);

		std::string st_info = conduct.type(path);
		if (st_info.empty())
		{
			if ("new" == action)
			{
				   if ((!wr(path,cmd)) || (conduct.type(path,false).empty()))
				   	out<<error("reason","failed");
			}
			else
				out<<error("reason","absent");

			continue;
		}

		if (st_info == "d")
		{
			if ("new" == action)
			{//ignore new for existent items
			}
			else if ("delete" == action)
			{
    				if (dir_level(path) == 2)
					conduct.del_plugin(path);
				else
				{
					conduct.destroy(path);
					if (!conduct.type(path).empty())
						out<<error("reason","failed");
				}
			}
			else if ("read" == action)
			{
				dirent_vector items;
				conduct.list(path,items);
				std::for_each(items.begin(),items.end(),out_cmd(out,cmd.name()));
			}
			else
				out<<error("reason","undefined");
		}
		else if (st_info == "r")
		{
			if ("new" == action)
			{//ignore new for nonexistent items
			}
			else if ("delete" == action)
			{
				conduct.destroy(path);
				if (!conduct.type(path).empty())
					out<<error("reason","failed");
			}
			else if ("read" == action)
			{
				command answer(cmd.name());
				out<<(rd(path,answer)?answer:error("reason","failed"));
			}
			else if ("write" == action)
			{
				if (!wr(path,cmd)) out<<error("reason","failed");
			}
			else
				out<<error("reason","undefined");
		}
		else
			out<<error("reason","unsupported");

	}
}


void bus_main(std::istream& is,std::ostream& os,std::ostream& log)
{
	while (is)
	{
		std::string direction;
		std::getline(is,direction);
		if (">>" == direction)
		{
			std::istringstream question(receive(is,os,log));
			std::ostringstream answer;
			forward(question,answer);
			send("<<",answer.str(),is,os,log);
		}
		else if ("<<" == direction)
		{
			os<<"<>"<<std::endl;
		}
		else
			log << "--invalid address line: \""
				<<direction << "\"" << std::endl;
	}
}


int main(int argc,char *argv[])
{
	for(;;)
	{
		static struct option long_options[] =
		{
			{"help", no_argument, 0, 'h'},
			{"version", no_argument, 0, 'v'},
			{"plugin-path", required_argument, 0, 'p'},
			{0, 0, 0, 0}
		};
		int c = getopt_long(argc, argv, "hvp:", long_options, NULL);
		if ( -1 == c) break;
		switch (c)
		{
			case 'h':
				usage(EXIT_SUCCESS);
			case 'v':
				std::cout << "defusator version " << VERSION << std::endl;
				std::cout << "Written by Stanislav Ievlev" << std::endl << std::endl;
				std::cout << "Copyright (C) 2004 ALT Linux Team" << std::endl;
				return (EXIT_SUCCESS);
			case 'p':
				plugin_dir = optarg;
				break;
			default:
				usage(EXIT_FAILURE);
		}
	}

	bus_main(std::cin,std::cout,std::cerr);
}
